@@ -3,7 +3,7 @@
 # Programmer friendly subprocess wrapper.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: November 8, 2015
+# Last Change: November 10, 2015
 # URL: https://executor.readthedocs.org
 
 """
@@ -41,6 +41,7 @@ Classes and functions
 """
 
 # Standard library modules.
+import errno
 import logging
 import os
 import pipes
@@ -50,6 +51,7 @@ import sys
 import tempfile
 
 # External dependencies.
+from humanfriendly import format
 from property_manager import PropertyManager, mutable_property, required_property, writable_property
 
 # Define an alias for Unicode strings that's unambiguous
@@ -62,7 +64,7 @@ except NameError:
     unicode = str
 
 # Semi-standard module versioning.
-__version__ = '7.2'
+__version__ = '7.6'
 
 # Initialize a logger.
 logger = logging.getLogger(__name__)
@@ -96,12 +98,18 @@ you're free to customize it if you really want to write your shell commands in
 ``fish`` or ``zsh`` syntax :-).
 """
 
+COMMAND_NOT_FOUND_STATUS = 127
+"""The exit status used by shells when a command is not found (an integer)."""
+
 DEFAULT_WORKING_DIRECTORY = os.curdir
 """
 The default working directory for external commands (a string). Defaults to the
 working directory of the current process using :data:`os.curdir`.
 """
 
+COMMAND_NOT_FOUND_CODES = (errno.ENOENT,)
+"""Numeric error codes returned when a command isn't available on the system (a tuple of integers)."""
+
 IS_WINDOWS = sys.platform.startswith('win')
 
 
@@ -449,17 +457,28 @@ class ExternalCommand(PropertyManager):
         """
         return {}
 
-    @property
+    @mutable_property
     def error_message(self):
         """A string describing how the external command failed or :data:`None`."""
-        if self.error_type is ExternalCommandFailed:
-            text = "External command failed with exit code %s! (command: %s)"
-            return text % (self.returncode, quote(self.command_line))
+        if self.error_type is CommandNotFound:
+            return format("External command isn't available! (command: %s, search path: %s)",
+                          quote(self.command_line), get_search_path())
+        elif self.error_type is ExternalCommandFailed:
+            return format("External command failed with exit code %s! (command: %s)",
+                          self.returncode, quote(self.command_line))
 
-    @property
+    @mutable_property
     def error_type(self):
-        """:exc:`ExternalCommandFailed` when :attr:`returncode` is a nonzero number, :data:`None` otherwise."""
-        if self.is_finished and self.failed:
+        """
+        An appropriate exception class or :data:`None` (when no error occurred).
+
+        :class:`CommandNotFound` if the external command exits with return code
+        :data:`COMMAND_NOT_FOUND_STATUS` or :exc:`ExternalCommandFailed` if the
+        external command exits with any other nonzero return code.
+        """
+        if self.returncode == COMMAND_NOT_FOUND_STATUS:
+            return CommandNotFound
+        elif self.returncode not in (None, 0):
             return ExternalCommandFailed
 
     @property
@@ -467,11 +486,14 @@ class ExternalCommand(PropertyManager):
         """
         Whether the external command has failed.
 
-        :data:`True` if :attr:`returncode` is a nonzero number, :data:`False`
-        if :attr:`returncode` is zero, :data:`None` when the external command
-        hasn't been started or is still running.
+        - :data:`True` if :attr:`returncode` is a nonzero number
+          or :attr:`error_type` is set (e.g. because the external
+          command doesn't exist).
+        - :data:`False` if :attr:`returncode` is zero.
+        - :data:`None` when the external command hasn't been started or is
+          still running.
         """
-        return (not self.succeeded) if self.is_finished else None
+        return (not self.succeeded) if self.succeeded is not None else None
 
     @mutable_property
     def fakeroot(self):
@@ -522,7 +544,7 @@ class ExternalCommand(PropertyManager):
         finished, :data:`False` when the external command hasn't been started
         yet or is still running.
         """
-        return self.subprocess.poll() is not None if self.subprocess else False
+        return self.error_type is not None or self.returncode is not None
 
     @property
     def is_running(self):
@@ -709,11 +731,17 @@ class ExternalCommand(PropertyManager):
         """
         Whether the external command succeeded.
 
-        :data:`True` if :attr:`returncode` is zero, :data:`False` if
-        :attr:`returncode` is a nonzero number, :data:`None` when the external
-        command hasn't been started or is still running.
+        - :data:`True` if :attr:`returncode` is zero.
+        - :data:`False` if :attr:`returncode` is a nonzero number
+          or :attr:`error_type` is set (e.g. because the external
+          command doesn't exist).
+        - :data:`None` when the external command hasn't been started or is
+          still running.
         """
-        return self.returncode == 0 if self.is_finished else None
+        if self.is_finished:
+            return self.returncode == 0
+        else:
+            return None
 
     @mutable_property
     def sudo(self):
@@ -742,7 +770,7 @@ class ExternalCommand(PropertyManager):
         .. _source shell command: https://en.wikipedia.org/wiki/Source_(command)
         """
 
-    @property
+    @mutable_property
     def was_started(self):
         """
         Whether the external command has already been started.
@@ -751,7 +779,7 @@ class ExternalCommand(PropertyManager):
         the external command, :data:`False` when :func:`start()` hasn't been
         called yet.
         """
-        return self.subprocess is not None
+        return self.error_type is not None or self.subprocess is not None
 
     def start(self):
         """
@@ -779,49 +807,43 @@ class ExternalCommand(PropertyManager):
         # Prepare the input.
         if self.input is not None:
             if self.async:
-                self.stdin_stream.prepare(self.encoded_input)
+                self.stdin_stream.prepare_input(self.encoded_input)
                 kw['stdin'] = self.stdin_stream.fd
             else:
                 kw['stdin'] = subprocess.PIPE
-        # Silence the standard output and error streams?
-        if self.silent:
+        # Prepare to capture the standard output and/or error stream(s).
+        kw['stdout'] = self.stdout_stream.prepare_output(self.stdout_file, self.capture, self.async)
+        # Make it possible to merge stderr into stdout.
+        kw['stderr'] = (subprocess.STDOUT if self.merge_streams else
+                        self.stderr_stream.prepare_output(self.stderr_file, self.capture_stderr, self.async))
+        # Silence the standard output and/or error stream(s)?
+        if self.silent and any(kw.get(n) is None for n in ('stdout', 'stderr')):
             if self.null_device is None:
                 self.null_device = open(os.devnull, 'wb')
-            kw['stdout'] = self.null_device
-            kw['stderr'] = self.null_device
-        # Prepare to capture the standard output stream.
-        if self.stdout_file:
-            self.stdout_stream.redirect(self.stdout_file)
-            kw['stdout'] = self.stdout_stream.fd
-        elif self.capture:
-            if self.async:
-                self.stdout_stream.prepare()
-                kw['stdout'] = self.stdout_stream.fd
-            else:
-                kw['stdout'] = subprocess.PIPE
-        # Make it possible to merge stderr into stdout.
-        if self.merge_streams:
-            kw['stderr'] = subprocess.STDOUT
-        elif self.stderr_file:
-            self.stderr_stream.redirect(self.stderr_file)
-            kw['stderr'] = self.stderr_stream.fd
-        elif self.capture_stderr:
-            # Prepare to capture the standard error stream.
-            if self.async:
-                self.stderr_stream.prepare()
-                kw['stderr'] = self.stderr_stream.fd
-            else:
-                kw['stderr'] = subprocess.PIPE
+            kw['stdout'] = self.null_device if kw['stdout'] is None else kw['stdout']
+            kw['stderr'] = self.null_device if kw['stderr'] is None else kw['stderr']
         # Create the subprocess object.
         self.logger.debug("Executing external command: %s", quote(kw['args']))
-        self.subprocess = subprocess.Popen(**kw)
+        # Clear previous values (if any).
+        delattr(self, 'error_type')
+        self.subprocess = None
+        try:
+            self.subprocess = subprocess.Popen(**kw)
+        except OSError as e:
+            if e.errno in COMMAND_NOT_FOUND_CODES:
+                # Enable uniform error handling.
+                self.error_type = CommandNotFound
+            else:
+                # Don't swallow exceptions we can't handle.
+                raise
         # Synchronously wait for the external command to end?
         if not self.async:
             # Feed the external command its input, capture the external
             # command's output, cleanup resources and check for errors.
-            stdout, stderr = self.subprocess.communicate(input=self.encoded_input)
-            self.stdout_stream.override(stdout)
-            self.stderr_stream.override(stderr)
+            if self.subprocess:
+                stdout, stderr = self.subprocess.communicate(input=self.encoded_input)
+                self.stdout_stream.override(stdout)
+                self.stderr_stream.override(stderr)
             self.wait()
 
     def wait(self, check=None):
@@ -850,7 +872,7 @@ class ExternalCommand(PropertyManager):
         """
         if not self.was_started:
             self.start()
-        if not self.is_finished:
+        if self.was_started and not self.is_finished:
             self.subprocess.wait()
         self.load_output()
         self.cleanup()
@@ -927,8 +949,24 @@ class ExternalCommand(PropertyManager):
             raise self.error_type(self)
 
     def __enter__(self):
-        """Start the external command if it hasn't already been started."""
+        """
+        Start the external command if it hasn't already been started.
+
+        :returns: The :class:`ExternalCommand` object.
+
+        When you use an :class:`ExternalCommand` as a context manager in the
+        :keyword:`with` statement, the command is automatically started when
+        entering the context and terminated when leaving the context.
+
+        If the proces hasn't already been started yet :attr:`async` is
+        automatically set to :data:`True` (if it's not already :data:`True`),
+        otherwise the command will have finished execution by the time the body
+        of the :keyword:`with` statement is executed (which isn't really all
+        that useful :-).
+        """
         if not self.was_started:
+            if not self.async:
+                self.async = True
             self.start()
         return self
 
@@ -968,6 +1006,47 @@ class CachedStream(object):
         self.is_temporary_file = False
         self.kind = kind
 
+    def prepare_temporary_file(self):
+        """Prepare the stream's temporary file."""
+        if not (self.fd and self.filename):
+            self.is_temporary_file = True
+            self.fd, self.filename = tempfile.mkstemp(prefix='executor-', suffix='-%s.txt' % self.kind)
+            logger.debug("Connected %s stream to temporary file %s ..", self.kind, self.filename)
+
+    def prepare_input(self, contents=None):
+        """
+        Initialize an asynchronous input stream (using a temporary file).
+
+        :param contents: If you pass this argument the given string will be
+                         written to the temporary file.
+        """
+        if contents is not None:
+            self.prepare_temporary_file()
+            with open(self.filename, 'wb') as handle:
+                handle.write(contents)
+
+    def prepare_output(self, file, capture, async):
+        """
+        Initialize an (asynchronous) output stream.
+
+        :param file: A file handle or :data:`None`.
+        :param capture: :data:`True` if capturing is enabled, :data:`False` otherwise.
+        :param async: :data:`True` for asynchronous execution, :data:`False` otherwise.
+        :returns: A file descriptor, :data:`subprocess.PIPE` or :data:`None`.
+        """
+        if file is not None:
+            # Capture the stream to a user defined file.
+            self.redirect(file)
+            return self.fd
+        elif capture:
+            if async:
+                # Capture the stream to a temporary file.
+                self.prepare_temporary_file()
+                return self.fd
+            else:
+                # Capture the stream in memory.
+                return subprocess.PIPE
+
     def redirect(self, obj):
         """
         Capture the stream in a file provided by the caller.
@@ -987,21 +1066,6 @@ class CachedStream(object):
             raise ValueError(msg % self.kind)
         logger.debug("Connected %s stream to file %s ..", self.kind, self.filename)
 
-    def prepare(self, contents=None):
-        """
-        Initialize the temporary file (and write the given string to it).
-
-        :param contents: If you pass this argument the given string will be
-                         written to the temporary file.
-        """
-        if not (self.fd and self.filename):
-            self.is_temporary_file = True
-            self.fd, self.filename = tempfile.mkstemp(prefix='executor-', suffix='-%s.txt' % self.kind)
-            logger.debug("Connected %s stream to temporary file %s ..", self.kind, self.filename)
-        if contents is not None:
-            with open(self.filename, 'wb') as handle:
-                handle.write(contents)
-
     def load(self):
         """
         Load the stream's contents from the temporary file.
@@ -1157,7 +1221,7 @@ def is_executable(filename, mode=os.F_OK | os.X_OK):
     return os.path.exists(filename) and os.access(filename, mode) and not os.path.isdir(filename)
 
 
-class ExternalCommandFailed(Exception):
+class ExternalCommandFailed(Exception, PropertyManager):
 
     """
     Raised when an external command exits with a nonzero status code.
@@ -1167,26 +1231,20 @@ class ExternalCommandFailed(Exception):
     an external command exits with a nonzero status code.
     """
 
-    def __init__(self, command, pool=None, error_message=None):
+    def __init__(self, command, **options):
         """
         Initialize an :class:`ExternalCommandFailed` object.
 
         :param command: The :class:`ExternalCommand` object that triggered the
                         exception.
-        :param pool: The :class:`.CommandPool` object that triggered the
-                     exception (optional).
+        :param kw: Keyword arguments are passed on to :func:`.PropertyManager.__init__()`.
         :param error_message: An error message to override the default message
                               taken from :attr:`~ExternalCommand.error_message`.
         """
-        # Initialize instance properties.
-        self.command = command
-        self.pool = None
-        if error_message:
-            self.error_message = error_message
-        # Initialize the superclass.
-        super(ExternalCommandFailed, self).__init__(self.error_message)
+        PropertyManager.__init__(self, command=command, **options)
+        Exception.__init__(self, self.error_message)
 
-    @writable_property(usage_notes=False)
+    @required_property(usage_notes=False)
     def command(self):
         """The :class:`ExternalCommand` object that triggered the exception."""
 
@@ -1214,3 +1272,28 @@ class ExternalCommandFailed(Exception):
         :func:`__init__()`.
         """
         return self.command.error_message
+
+
+class CommandNotFound(ExternalCommandFailed, OSError):
+
+    """
+    Raised when an external command is not available on the system.
+
+    This exception is raised by :func:`execute()`,
+    :func:`~ExternalCommand.start()` and :func:`~ExternalCommand.wait()` when
+    an external command can't be started because the command isn't available.
+
+    It inherits from :exc:`ExternalCommandFailed` to enable uniform error
+    handling but it also inherits from :exc:`~exceptions.OSError` for
+    backwards compatibility (see :attr:`errno` and :attr:`strerror`).
+    """
+
+    @property
+    def errno(self):
+        """The numeric error code :data:`~errno.ENOENT` from :mod:`errno` (an integer)."""
+        return errno.ENOENT
+
+    @property
+    def strerror(self):
+        """The text corresponding to :attr:`errno` (a string)."""
+        return os.strerror(self.errno)
